/* SPARC code for booting linux 2.6
 *
 * (C) Copyright 2007
 * Daniel Hellstrom, Gaisler Research, daniel@gaisler.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <asm/byteorder.h>
#include <asm/prom.h>
#include <asm/cache.h>

#define PRINT_KERNEL_HEADER

extern image_header_t header;
extern void srmmu_init_cpu(unsigned int entry);
extern void prepare_bootargs(char *bootargs);

/* sparc kernel argument (the ROM vector) */
struct linux_romvec *kernel_arg_promvec;

/* page szie is 4k */
#define PAGE_SIZE 0x1000
#define RAMDISK_IMAGE_START_MASK	0x07FF
#define RAMDISK_PROMPT_FLAG		0x8000
#define RAMDISK_LOAD_FLAG		0x4000
struct __attribute__((packed)) {
	char traptable[PAGE_SIZE];
	char swapper_pg_dir[PAGE_SIZE];
	char pg0[PAGE_SIZE];
	char pg1[PAGE_SIZE];
	char pg2[PAGE_SIZE];
	char pg3[PAGE_SIZE];
	char empty_bad_page[PAGE_SIZE];
	char empty_bad_page_table[PAGE_SIZE];
	char empty_zero_page[PAGE_SIZE];
	unsigned char hdr[4]; /* ascii "HdrS" */
	/* 00.02.06.0b is for Linux kernel 2.6.11 */
	unsigned char linuxver_mega_major; 
	unsigned char linuxver_major;  
	unsigned char linuxver_minor;
	unsigned char linuxver_revision;
	/* header version 0x0203 */
	unsigned short hdr_ver;
	union __attribute__((packed)) {
		struct __attribute__((packed)) {
			unsigned short root_flags;
			unsigned short root_dev;
			unsigned short ram_flags;
			unsigned int sparc_ramdisk_image;
			unsigned int sparc_ramdisk_size;
			unsigned int reboot_command;
			unsigned int resv[3];
			unsigned int end;
		} ver_0203;
	} hdr_input;
} *linux_hdr;

/* temporary initrd image holder */
image_header_t ihdr;

/* boot the linux kernel */
void do_bootm_linux(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[],
		ulong addr, ulong *len_ptr, int   verify)
{
	image_header_t *hdr = &header;
	
	char *bootargs;
	unsigned int data, len, checksum;
	unsigned int initrd_addr, kernend;
	void (*kernel)(struct linux_romvec *, void *) = 
	  (void (*)(struct linux_romvec *romvec, void *dbg)) ntohl (hdr->ih_ep);
	
	/* Get virtual address of kernel start */
	linux_hdr = (unsigned int) ntohl(hdr->ih_load);
	
	/* check for a SPARC kernel */
	if ( (linux_hdr->hdr[0] != 'H') ||
	     (linux_hdr->hdr[1] != 'd') ||
	     (linux_hdr->hdr[2] != 'r') ||
	     (linux_hdr->hdr[3] != 'S') ){
	     	printf("Error reading header of SPARC Linux kernel, aborting\n");
		return;
	}
	
#ifdef PRINT_KERNEL_HEADER 
	printf("## Found SPARC Linux kernel %d.%d.%d.%d ...\n",
	  linux_hdr->linuxver_mega_major,
	  linux_hdr->linuxver_major,
	  linux_hdr->linuxver_minor,
	  linux_hdr->linuxver_revision);
#endif
	/* set basic boot params in kernel header now that it has been 
	 * extracted and is writeable.
	 */
	 
	/*
	 * Are we going to use an initrd image
	 */
	if (argc >= 3) {
		initrd_addr = simple_strtoul (argv[2], NULL, 16);

		printf ("## Loading Ramdisk Image at %08lx ...\n", initrd_addr);
	
		/* Copy header so we can blank CRC field for re-calculation */
		memmove (&ihdr, (char *)initrd_addr, sizeof (image_header_t));
		
		if (ntohl (ihdr.ih_magic) != IH_MAGIC) {
			printf ("Bad Magic Number\n");
			do_reset (cmdtp, flag, argc, argv);
		}
		
		data = (ulong) &ihdr;
		len = sizeof (image_header_t);

		checksum = ntohl (ihdr.ih_hcrc);
		ihdr.ih_hcrc = 0;
		
		/* Always verify Header Checksum */
		if (crc32 (0, (unsigned char *) data, len) != checksum) {
			printf ("Bad Header Checksum\n");
			do_reset (cmdtp, flag, argc, argv);
		}
		
		print_image_hdr(&ihdr);

		data = initrd_addr + sizeof (image_header_t);
		len = ntohl (ihdr.ih_size);
		
		/* Cannot run Initrd image from FLASH.
		 * We move initrd image just after kernel.
		 * We can get kernel end by reading kernel 
		 * header.
		 */
		kernend = linux_hdr->hdr_input.ver_0203.end;
		/* Virtual to physical address */
		kernend = (kernend & 0x0fffffff) | ((unsigned int)linux_hdr&0xf0000000);
		/* Align to Page */
		kernend = (kernend + PAGE_SIZE) & ~(PAGE_SIZE-1);

		printf ("   Loading Ramdisk to %08lx, end %08lx ... ",
			kernend, kernend+len);
		
		/* Copy from flash/ram to ram after kernel */	
		memmove((void *)kernend,(void *)data,len);
		puts("Done\n");
		
		/* Work with new initrd image */
		data = kernend;
		
		/* Verify DATA in image using crc32 */
		if (verify) {
			ulong csum = 0;

			printf ("   Verifying Checksum ... ");
			csum = crc32(0, (unsigned char *) data, len);
			if (csum != ntohl (ihdr.ih_dcrc)) {
				printf ("Bad Data CRC\n");
				do_reset (cmdtp, flag, argc, argv);
			}else{
				printf ("OK\n");
			}
		}


		if ((ihdr.ih_os != IH_OS_LINUX) ||
		    (ihdr.ih_arch != IH_CPU_SPARC) ||
		    (ihdr.ih_type != IH_TYPE_RAMDISK)) {
			printf ("No Linux SPARC Ramdisk Image, (%d,%d,%d)\n",
			  ihdr.ih_os,ihdr.ih_arch,ihdr.ih_type);
			do_reset (cmdtp, flag, argc, argv);
		}
		
		/* 
		 * Set INITRD Image address relative to RAM Start
		 */
		linux_hdr->hdr_input.ver_0203.sparc_ramdisk_image = 
		  data - CFG_SDRAM_BASE;
		linux_hdr->hdr_input.ver_0203.sparc_ramdisk_size = len;
		/* Clear READ ONLY flag if set to non-zero */
		linux_hdr->hdr_input.ver_0203.root_flags = 1;
		/* Set root device to: Root_RAM0 */
		linux_hdr->hdr_input.ver_0203.root_dev = 0x100;
		linux_hdr->hdr_input.ver_0203.ram_flags = 0;
	}else{
		/* NOT using RAMDISK image, overwriting kernel defaults */
		linux_hdr->hdr_input.ver_0203.sparc_ramdisk_image = 0;
		linux_hdr->hdr_input.ver_0203.sparc_ramdisk_size = 0;
		/* Leave to kernel defaults
		linux_hdr->hdr_input.ver_0203.root_flags = 1;
		linux_hdr->hdr_input.ver_0203.root_dev = 0;
		linux_hdr->hdr_input.ver_0203.ram_flags = 0;
		*/
	}
	
	/* Copy bootargs from bootargs variable to kernel readable area */
	bootargs = getenv("bootargs");
	prepare_bootargs(bootargs);
	
	/* turn on mmu & setup context table & page table for process 0 (kernel) */
	srmmu_init_cpu((unsigned int)kernel);

	/* Enter SPARC Linux kernel
	 * From now on the only code in u-boot that will be
	 * executed is the PROM code.
	 */
	kernel(kernel_arg_promvec, (void *)hdr->ih_ep);
	
	/* It will never come to this... */
	while(1) ;
}
